#!/usr/bin/env bash

PS_ROOT="/home/pinsound/"
PS_AUDIO="${PS_ROOT}audio/"
PS_USB_MNT="/tmp/pinsound/"
PS_DEV_USB="/dev/sda1"

PS_PLAYPINSOUND="${PS_ROOT}bin/playPinSound"

PS_ALSA_VOLUME_BASH="70%"
PS_ALSA_VOLUME_PLAYPINSOUND="100%"

function endUpdate()
{
	mount -o remount,ro /

	freeAudio
	aplay ${PS_AUDIO}The_system_is_now_updated.wav
	for run in {1..3}
	do
		aplay ${PS_AUDIO}Please__restart_the_pinball_machine_now.wav
	done
	exit 0
}

function checkUpdate()
{
	if [ -f ${PS_USB_MNT}pinsound-init.bin ] || [ -f ${PS_USB_MNT}pinsound-init.psrom ]
	then
	        cp ${PS_USB_MNT}pinsound-init.* ${PS_USB_MNT}pinsound-update-0000.psrom
	        sync
	fi

	if ls ${PS_USB_MNT}pinsound-update*.bin &> /dev/null
	then
		UPDATE_FILENAME_BIN=`ls ${PS_USB_MNT}pinsound-update*.bin | tail -1`
		echo "Found pinsound-update BIN -> conver to PSROM"
		mv "${UPDATE_FILENAME_BIN}" "${UPDATE_FILENAME_BIN%.bin}.psrom"
	fi

	if ls ${PS_USB_MNT}pinsound-update*.psrom &> /dev/null
	then
		UPDATE_FILENAME=`ls ${PS_USB_MNT}pinsound-update*.psrom | tail -1`

		freeAudio
		aplay ${PS_AUDIO}The_system_is_currently_updating._Please_wait.wav
		mpg123 -q --loop 99 ${PS_ROOT}audio/music_updating.mp3 &

		# to be sure, remount root filesystem in RW
		mount -o remount,rw /

		if  [ ! -e /tmp/update/ ]  
		then
			mkdir /tmp/update/
		fi
		cp ${UPDATE_FILENAME} /tmp/update/
		UPDATE_FILENAME_BASENAME=$(basename $UPDATE_FILENAME)
		mv /tmp/update/${UPDATE_FILENAME_BASENAME} /tmp/update/update.tar.gz
		# remove old previous master configs
		rm -f ${PS_ROOT}pinball/*
		tar zxf /tmp/update/update.tar.gz -C /
		rm /tmp/update/update.tar.gz
		rm ${UPDATE_FILENAME}
		sync

		# check update STM
		checkUpdateStm

		endUpdate
	fi
	return 1
}

function checkUpdateStm()
{
	
	if ls ${PS_ROOT}update/firmware*.bin &> /dev/null
	then
		
		FIRMWARE_FILENAME=`ls ${PS_ROOT}/update/firmware*.bin | tail -1`
		FIRMWARE_VERSION=${FIRMWARE_FILENAME#*-}; # FIRMWARE_VERSION = "-X.bin"
		FIRMWARE_VERSION=${FIRMWARE_VERSION%-*}; 	# FIRMWARE_VERSION = "X"

		# get current STM firmware version
		# disabled
		#${PS_PLAYPINSOUND} --comparefirmware ${FIRMWARE_VERSION}
		#STM_TO_UPDATE=$?

		STM_TO_UPDATE=1
		
		if [[ "$STM_TO_UPDATE" == 1 ]]
		then
			
			for run in {1..3}
			do
				${PS_ROOT}bin/stm32flash /dev/ttySP1 -w  ${FIRMWARE_FILENAME} -i 92,-91,91:-92,-91 -b 115200
				funcResult=$?
				if [[ $funcResult -eq 0 ]] 
				then
					rm ${PS_ROOT}/update/firmware*.bin
					sync
					return 0
				fi
			done

			if [[ $funcResult -eq 1 ]] 
			then
				freeAudio
				aplay ${PS_AUDIO}Technical_problem_please_contact_the_support_team.wav
				exit 1
			fi
		fi
		rm ${PS_ROOT}/update/firmware*.bin
		sync
	fi
	return 1
}

function checkInstallPrerequisites()
{
	# check if binary playPinSound is available
	if [ ! -f ${PS_PLAYPINSOUND} ]
	then
		return 0
	fi

	# check if "audio" directory is created on the USB flash drive
	if [ ! -d ${PS_USB_MNT}audio ]
	then
		mkdir ${PS_USB_MNT}audio
	fi
	
	# check if "tmp" directory is created on the USB flash drive
	if [ ! -d ${PS_USB_MNT}tmp ]
	then
		mkdir ${PS_USB_MNT}tmp
	fi

	# are tools "oggdec" and "unzip" available?
	which oggdec unzip dosfsck
	result_which=$?
	if [[ $result_which != 0 ]] 
	then
		mount -o remount,rw /
		pacman --noconfirm -U ${PS_ROOT}pkg/libao-1.2.0-1-arm.pkg.tar.xz ${PS_ROOT}pkg/vorbis-tools-1.4.0-4-arm.pkg.tar.xz ${PS_ROOT}pkg/unzip-6.0-7-arm.pkg.tar.xz ${PS_ROOT}pkg/dosfstools-3.0.26-1-arm.pkg.tar.xz
		result_pacman=$?
		mount -o remount,ro /
		if [[ $result_pacman == 1 ]] 
		then
			freeAudio
			aplay ${PS_ROOT}audio/Due_to_major_problems__please_update_the_PinSound_system.wav
			echo "missing pkg archives"
			return 1;
		fi
	fi

	# check if STM is here
	# DISABLED ! Integrated inside playPinsound at startup
	#${PS_PLAYPINSOUND} --pingfirmware
	#STM_TO_UPDATE=$?
	#if [[ "$STM_TO_UPDATE" == 0 ]]
	#then
	#	mount -o remount,rw /
	#	tar zxf ${PS_ROOT}default/default.tar.gz -C / --wildcards "*.bin"  
	#	sync
	#	checkUpdateStm
	#	endUpdate
	#fi

	return 1
}

function loadDefault()
{
	# to be sure, remount root filesystem in RW
	mount -o remount,rw /

	if [ ! -f ${PS_ROOT}default/default.tar.gz ]
	then
		tar zxf ${PS_ROOT}default/default.tar.gz -C /
		sync
		checkUpdateStm
		endUpdate
	fi
	return 0
}

function checkDiskUsb()
{
	freeAudio
	aplay ${PS_ROOT}audio/Checking_USB_flash_drive_for_errors_please_wait.wav
	mpg123 -q --loop 99 ${PS_ROOT}audio/music_updating.mp3 &
	
	checkInstallPrerequisites

	# unmount USB
	umount ${PS_DEV_USB}

	# launch dosfsck
	dosfsck -a -w ${PS_DEV_USB}
	result_fsck=$?

	# remount USB
	mountpoint -q ${PS_USB_MNT} || mount -o rw ${PS_DEV_USB} ${PS_USB_MNT}

	# 1 = errors found
	# 0 = no errors
	return $result_fsck
}

function processTmp()
{
	# clean hidden files (thank you OSX with your fucking hidden files...)
	find "${PS_USB_MNT}tmp" -name '.*' ! -name '.' ! -name '..' -delete

	find ${PS_USB_MNT}tmp -maxdepth 1 -type d ! -path ${PS_USB_MNT}tmp | while read mix_dir; do
		
		has_ogg=$(hasOggFiles "$mix_dir")
		if [[ $has_ogg -eq 0 ]]
		then	
			echo "Mix $mix_dir contains no OGG files"

			# wav files, we can directly move this mix to audio directory
			mv "$mix_dir" ${PS_USB_MNT}audio
			sync
			echo "Mix $mix_dir moved to audio folder"

			freeAudio
			aplay ${PS_ROOT}audio/completed.wav
			
			# launch only one conversion at boot
			break
		fi

		# get count of OGG files to convert
		total_ogg=$(find "$mix_dir" -type f -iname "*.ogg" | wc -l)
		echo "total OGG : $total_ogg"

		# launch ogg to wav conversion
		random_music=$[($RANDOM % ($[3 - 1] + 1)) + 1]
		freeAudio
		mpg123 -q --loop 99 ${PS_ROOT}audio/music_conversion_${random_music}.mp3 &
		echo "Ogg to Wav conversion : $mix_dir"
		count=0
		current_percent=-1
		find "$mix_dir" -type f -iname "*.ogg" | sort -R | while read ogg_file; do
			
			#wav_file="$(dirname $ogg_file)/$(basename $ogg_file .ogg).wav"
			#echo "$ogg_tmp => $wav_file"
			oggdec -Q "$ogg_file"
			result_oggdec=$?
			if [[ $result_oggdec == 0 ]] 
			then
				rm "$ogg_file"
				sync
			else
				sync

				echo "Error during wav conversion"
				freeAudio
				for run in {1..2}
				do
					aplay ${PS_ROOT}audio/Error_encountered_during_audio_conversion.wav
				done

				# maybe it's a disk space problem?
				usb_free_space=$(getFreeSpace ${PS_USB_MNT})
				if [[ $usb_free_space -lt 50000000 ]]; then
					echo "The USB flash drive does not have enough free space."
					for run in {1..3}
					do
						aplay ${PS_ROOT}audio/The_USB_flash_drive_does_not_have_enough_free_space.wav
					done
				fi

				# launch dosfsck to check the USB flash drive
				checkDiskUsb

				# user must fix the problem, we exit
				return 1;
			fi
			
			count=$((count+1))
			#echo -ne "Progression : $count / $total_ogg \r"
			percent_progress=$(($count * 100 / $total_ogg))
			if [[ $percent_progress != $current_percent ]]; then
				if [[ $(expr $current_percent % 10) == 0 ]]; then
					if [ -f ${PS_ROOT}audio/${current_percent}_percent_completed.wav ]; then
						freeAudio
						aplay ${PS_ROOT}audio/${current_percent}_percent_completed.wav
						skip_frames=$[($RANDOM % ($[4000 - 500] + 1)) + 500]
						mpg123 -q --loop 99 --skip $skip_frames ${PS_ROOT}audio/music_conversion_${random_music}.mp3 &
					fi
				fi
				echo "Progression : $count / $total_ogg | Percent : $percent_progress"
			else
				echo "."
			fi
			current_percent=$percent_progress
			#sleep 0.2

		done
		
		has_ogg=$(hasOggFiles "$mix_dir")
		if [[ $has_ogg -eq 0 ]]
		then	
			# no more OGG files, we can move this mix to audio directory
			mv "$mix_dir" ${PS_USB_MNT}audio
			sync

			freeAudio
			aplay ${PS_ROOT}audio/completed.wav
		else
			echo "Error during wav conversion : still OGG files !"
			freeAudio
		fi

		# launch only one OGGDEC conversion at boot
		break
	done

	freeAudio
	return 0;
}

function freeAudio()
{
	killall mpg123 &> /dev/null
	killall aplay &> /dev/null
}

function findUnzipAudioMixZip()
{
	find ${PS_USB_MNT} -maxdepth 2 -type f -iname "*.zip" -a ! -iname "ERROR-*" | grep -v "${PS_USB_MNT}exclude/" | grep -v "${PS_USB_MNT}tmp/" | while read zip_file; do
		
		#check space available on USB flash drive
		usb_free_space=$(getFreeSpace ${PS_USB_MNT})
		zip_size=$(stat -c%s "$zip_file")
		echo "$zip_file ~ size : $zip_size / $usb_free_space"
		unzip_space=$(echo "$zip_size * 2")
		if [[ $usb_free_space -gt $unzip_space ]]
		then

			random_music=$[($RANDOM % ($[3 - 1] + 1)) + 1]
			freeAudio
			mpg123 -q --loop 99 ${PS_ROOT}audio/music_conversion_${random_music}.mp3 &

			zip_tmp_path="${PS_USB_MNT}tmp/$(basename "$zip_file")"

			if [ -f "${zip_tmp_path}" ]
			then
			  rm "${zip_tmp_path}"
			  sync
			fi
			mv "${zip_file}" "${PS_USB_MNT}tmp/"
			sync
			
			#### DISABLED !!! ####		
			if [ 1 -eq 0 ]; then

				# first, we test the zip archive
				unzip -q -t "$zip_tmp_path"
				result_unzip=$?
				if [[ $result_unzip != 0 ]] 
				then
					# we move zip file to corrupted
					mv "$zip_tmp_path" "${PS_USB_MNT}ERROR-$(basename "$zip_file")"
					sync

					echo "zip may be corrupted"
					freeAudio
					for run in {1..3}
					do
						aplay ${PS_ROOT}audio/Zip_file_may_be_corrupted.wav
					done

					# user must fix the problem, we exit
					return 1;
				fi

			fi
			#### END DISABLED !!! ####

			# we unzip to tmp dir
			echo "> Unzip $zip_tmp_path"
			unzip -q -o -d ${PS_USB_MNT}tmp "${zip_tmp_path}"
			result_unzip=$?
			if [[ $result_unzip != 0 ]] 
			then
				sync
				echo "error during zip extraction"
				freeAudio
				for run in {1..3}
				do
					aplay ${PS_ROOT}audio/Error_during_zip_extraction.wav
				done

				return 1;
			else
				# remove zip file, useless now
				rm "$zip_tmp_path"
				sync
			fi

			#check if unzipped mix is without directory container
			without_container=$(checkMoveMixWithoutContainer ${PS_USB_MNT}tmp ${PS_USB_MNT}tmp)

		else
			echo "The USB flash drive does not have enough free space."
			freeAudio
			for run in {1..3}
			do
				aplay ${PS_ROOT}audio/The_USB_flash_drive_does_not_have_enough_free_space.wav
			done
			return 1;
		fi

	done;

	return 0;
}

function findMoveAudioMix()
{
	if [[ $# -eq 0 ]]; then
		search_root=${PS_USB_MNT}
	else
		search_root="${1}"
	fi

	current_mix_directory=""

	# try to find if there are audio mixes elsewhere from /audio
	# everything is scanned, except mix in "exclude" directory
	echo ${PS_USB_MNT}exclude
	find "$search_root" -maxdepth 2 -type d \( ! -path '*/.*' \) -a \( ! -path "${PS_USB_MNT}exclude/*" \) -a \( ! -path "${PS_USB_MNT}tmp/*" \) -a \( -name "music" -o -name "sfx" -o -name "voice" \) | sed 's,/*[^/]\+/*$,,' | sort | while read mix_directory; do
    	if [ "$current_mix_directory" != "$mix_directory" ]; then
    		echo "*** $mix_directory"
			is_wav=$(isMixWavFormat "$mix_directory")
			if [[ $is_wav -eq 1 ]]
			then			
				moveAudioMix "$mix_directory" "audio"			
			else
				moveAudioMix "$mix_directory" "tmp"
			fi
    	fi
    	current_mix_directory="$mix_directory"
	done
}

function moveAudioMix()
{
	if [ ! -d ${PS_USB_MNT}$2 ]
	then
		mkdir ${PS_USB_MNT}$2
	fi

	directory=$(basename $1)

	if [ "$1" == "${PS_USB_MNT}audio" ] || [ "$1/" == "${PS_USB_MNT}" ]
	then
		# specific case : mix without container music,sfx,voice ...
		checkMoveMixWithoutContainer $1 ${PS_USB_MNT} 
		move_result=$?
		if [[ $move_result -eq 1 ]] 
		then
			return;
		fi
	fi

	# check if we try to move directory already at the right place
	if [ "${PS_USB_MNT}$2/$directory" != "$1" ]
	then
		echo "Move $1 mix into $2 : "
		mv "$1" "${PS_USB_MNT}$2"
		sync
		return;
	fi

	echo "Nothing to do"
	return;
}

function checkMoveMixWithoutContainer()
{
	count_types=`find "$1" \( ! -path '*/.*' \) -maxdepth 1 -type d \( -name "music" -o -name "sfx" -o -name "voice" -o -name "single" -o -name "jingle" \) 2>/dev/null | wc -l`
	if [[ $count_types -gt 0 ]]
	then
		# create new mix directory
		is_ogg=$(hasOggFiles "$1")
		if [[ $is_ogg -eq 1 ]]
		then			
			new_mix_dir=$(mktemp -d --tmpdir="$2/tmp" pinsound_mix_XXXX)
		else
			new_mix_dir=$(mktemp -d --tmpdir="$2/audio" pinsound_mix_XXXX)		
		fi
		
		find "${1}" \( ! -path '*/.*' \) -maxdepth 1 -type d \( -name "music" -o -name "sfx" -o -name "voice" -o -name "single" -o -name "jingle" \) | while read type_dir; do
			echo "$type_dir to mix : $new_mix_dir"
			mv "$type_dir" "$new_mix_dir"
			sync
		done
		return 1
	fi
	return 0
}

function isMixWavFormat()
{
	count_wav=`find "$1" -type f -iname "*wav" 2>/dev/null | wc -l`
	if [[ $count_wav -eq 0 ]]
	then
		echo 0
	else
		echo 1
	fi
}

function hasOggFiles()
{
	count_ogg=`find "$1" -type f -iname "*ogg" 2>/dev/null | wc -l`
	if [[ $count_ogg -gt 0 ]]
	then
		echo 1
	else
		echo 0
	fi
}

function getFreeSpace()
{
	free_space=$(df -P -B1 $1 | tail -1 | awk '{print $4}')
	echo $free_space
}

# rename PinSound directories to lower case 
function renamePinSoundDirectories()
{
#	find ${PS_USB_MNT} -maxdepth 1 -type d \( -iname "audio" -o -iname "playback" -o -iname "exclude" -o -iname "tmp" \)  -exec rename 's/(.*)\/([^\/]*)/$1\/\L$2/' {} \;
	return 0;
}

#########################################################
#               MAIN PROCESS
#########################################################

# set ALSA volume not too high!
amixer set DAC $PS_ALSA_VOLUME_BASH

# blink led, linux system is running 
${PS_ROOT}bin/ledBlink &

# needed to initialize alsa & stereo
aplay ${PS_AUDIO}empty.wav

# mount pinsound usb
if  [ ! -e ${PS_USB_MNT} ]  
then
	mkdir ${PS_USB_MNT}
fi

if  [ -e ${PS_DEV_USB} ]  
then
	mountpoint -q ${PS_USB_MNT} || mount -o rw ${PS_DEV_USB} ${PS_USB_MNT}

	# check debug mode
	if [ -f ${PS_USB_MNT}config.ini ] && [ -f ${PS_ROOT}bin/readini ]
	then
		debug_boot=$(${PS_ROOT}bin/readini ${PS_USB_MNT}config.ini main debug_boot)
		if [[ $debug_boot -eq 1 ]]
		then
			exec 1> ${PS_USB_MNT}pinsound-boot-debug.log 2>&1
			echo "PinSound boot log enabled"
		else
			if [ -f ${PS_USB_MNT}pinsound-boot-debug.log ]
			then
				rm ${PS_USB_MNT}pinsound-boot-debug.log
			fi
		fi
	fi

	checkUpdate
	funcResult=$?
	if [[ $funcResult -eq 0 ]] 
	then
		exit 1
	fi

	checkInstallPrerequisites
	funcResult=$?
	if [[ $funcResult -eq 0 ]] 
	then
		# if we got here, binaries are missing, we load default binaries
		#loadDefault

		# we recheck again
		checkInstallPrerequisites
		funcResult2=$?
		if [[ $funcResult2 -eq 0 ]] 
		then
			freeAudio
			aplay ${PS_AUDIO}Technical_problem_please_contact_the_support_team.wav
			exit 1
		fi
	fi

	mount -o remount,ro /

	# manage audio mixes
	renamePinSoundDirectories;
	findMoveAudioMix;
	findUnzipAudioMixZip;
	processTmp;
	

	# launch pinsound software
	amixer set DAC $PS_ALSA_VOLUME_PLAYPINSOUND
	for run in {1..3}
	do
		echo "Launch PinSound Software..."
  		${PS_PLAYPINSOUND} ${PS_USB_MNT}
	done
	funcResult=$?

	amixer set DAC $PS_ALSA_VOLUME_BASH

	# try to find what is the problem

	# exit 10 = no sound package available on the USB flash drive
	if [[ $funcResult -eq 10 ]] 
	then
		echo "Error : no sound files on the USB flash drive or no sound package"
		for run in {1..3}
		do
			freeAudio
			aplay ${PS_AUDIO}There_are_no_sound_files_on_the_USB_flash_drive.wav
		done
		exit 1
	fi

	if [ ! -d ${PS_USB_MNT}audio ]
	then
		echo "Error : no audio directory"
		mkdir ${PS_USB_MNT}audio
		touch ${PS_USB_MNT}audio/put_your_audio_mixes_here.txt
		for run in {1..3}
		do
			freeAudio
			aplay ${PS_AUDIO}There_are_no_sound_files_on_the_USB_flash_drive.wav
		done
		exit 1
	fi

	if ! test "$(ls -A "${PS_USB_MNT}audio/")"
	then
		echo "Error : no sound files on the USB flash drive"
		for run in {1..3}
		do
			freeAudio
			aplay ${PS_AUDIO}There_are_no_sound_files_on_the_USB_flash_drive.wav
		done
		touch ${PS_USB_MNT}audio/put_your_audio_mixes_here.txt
		exit 1
	fi

	count_wav=`find ${PS_USB_MNT}audio -type f -iname "*wav" 2>/dev/null | wc -l`
	if [[ $count_wav -eq 0 ]]
	then
		echo "Error : no wav files in the audio directory"
		for run in {1..3}
		do
			freeAudio
			aplay ${PS_AUDIO}There_are_no_wave_files_in_the_audio_directory.wav
		done
		exit 1
	fi

	# last chance, we check the USB flash drive
	checkDiskUsb
	result_checkdisk=$?
	if [[ $result_checkdisk != 0 ]]
	then
		echo "Errors found on the USB flash drive"
		freeAudio
		for run in {1..3}
		do
			freeAudio
			aplay ${PS_AUDIO}Please__restart_the_pinball_machine_now.wav
		done
		exit 0
	fi


	# if we get here, there is a major problem
	freeAudio
	aplay ${PS_AUDIO}Technical_problem_please_contact_the_support_team.wav
	exit 1

else
	freeAudio
	aplay ${PS_AUDIO}The_USB_flash_drive_is_not_present.wav
	exit 1
fi


